# stdlib
import ast
from collections.abc import Callable
import inspect
from inspect import Signature
import keyword
import linecache
import re
import textwrap
from textwrap import dedent
from typing import Any
from typing import cast

# third party
from pydantic import ValidationError
from pydantic import field_validator
from pydantic import model_validator

# relative
from ...abstract_server import AbstractServer
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...serde.signature import signature_remove_context
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_migration import migrate
from ...types.syft_object import PartialSyftObject
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftObject
from ...types.syncable_object import SyncableSyftObject
from ...types.transforms import TransformContext
from ...types.transforms import drop
from ...types.transforms import generate_action_object_id
from ...types.transforms import generate_id
from ...types.transforms import keep
from ...types.transforms import make_set_default
from ...types.transforms import transform
from ...types.uid import UID
from ...util.misc_objs import MarkdownDescription
from ..context import AuthedServiceContext
from ..response import SyftError
from ..user.user import UserView
from .utils import print as log_print

NOT_ACCESSIBLE_STRING = "N / A"


class HelperFunctionSet:
    def __init__(self, helper_functions: dict[str, Callable]) -> None:
        self.helper_functions = helper_functions
        for name, func in helper_functions.items():
            setattr(self, name, func)


class TwinAPIAuthedContext(AuthedServiceContext):
    __canonical_name__ = "TwinAPIAuthedContext"
    __version__ = SYFT_OBJECT_VERSION_1

    user: UserView | None = None
    settings: dict[str, Any] | None = None
    code: HelperFunctionSet | None = None
    state: dict[Any, Any] | None = None
    admin_client: SyftClient | None = None
    user_client: SyftClient | None = None


@serializable()
class TwinAPIContextView(SyftObject):
    __canonical_name__ = "TwinAPIContextView"
    __version__ = SYFT_OBJECT_VERSION_1

    __repr_attrs__ = ["settings", "state", "user"]
    user: UserView
    settings: dict[str, Any]
    state: dict[Any, Any]


def get_signature(func: Callable) -> Signature:
    sig = inspect.signature(func)
    sig = signature_remove_context(sig)
    return sig


def register_fn_in_linecache(fname: str, src: str) -> None:
    """adds a function to linecache, such that inspect.getsource works for functions nested in this function.
    This only works if the same function is compiled under the same filename"""
    lines = [
        line + "\n" for line in src.splitlines()
    ]  # use same splitting method same as linecache 112 (py3.12)
    linecache.cache[fname] = (137, None, lines, fname)


@serializable()
class TwinAPIEndpointViewV1(SyftObject):
    # version
    __canonical_name__ = "CustomAPIView"
    __version__ = SYFT_OBJECT_VERSION_1

    path: str
    action_object_id: UID
    signature: Signature
    access: str = "Public"
    mock_function: str | None = None
    private_function: str | None = None
    description: MarkdownDescription | None = None
    mock_helper_functions: list[str] | None = None
    private_helper_functions: list[str] | None = None
    worker_pool: str | None = None
    endpoint_timeout: int = 60


@serializable()
class TwinAPIEndpointView(SyftObject):
    # version
    __canonical_name__ = "CustomAPIView"
    __version__ = SYFT_OBJECT_VERSION_2

    path: str
    action_object_id: UID
    signature: Signature
    access: str = "Public"
    mock_function: str | None = None
    private_function: str | None = None
    description: MarkdownDescription | None = None
    mock_helper_functions: list[str] | None = None
    private_helper_functions: list[str] | None = None
    worker_pool_name: str | None = None
    endpoint_timeout: int = 60

    __repr_attrs__ = [
        "path",
        "signature",
        "worker_pool_name",
        "endpoint_timeout",
    ]

    def _coll_repr_(self) -> dict[str, Any]:
        if self.mock_function:
            mock_parsed_code = ast.parse(self.mock_function)
            mock_function_name = [
                server.name
                for server in ast.walk(mock_parsed_code)
                if isinstance(server, ast.FunctionDef)
            ][0]
        else:
            mock_function_name = NOT_ACCESSIBLE_STRING

        if self.private_function:
            private_parsed_code = ast.parse(self.private_function)
            private_function_name = [
                server.name
                for server in ast.walk(private_parsed_code)
                if isinstance(server, ast.FunctionDef)
            ][0]
        else:
            private_function_name = NOT_ACCESSIBLE_STRING

        worker_pool_name = "UNSET (DEFAULT)"
        if self.worker_pool_name is not None:
            worker_pool_name = self.worker_pool_name
        return {
            "API path": self.path,
            "Signature": self.path + str(self.signature),
            "Access": self.access,
            "Mock Function": mock_function_name,
            "Private Function": private_function_name,
            "Worker Pool": worker_pool_name,
        }


@serializable()
class Endpoint(SyftObject):
    """Base class to perform basic Endpoint validation for both public/private endpoints."""

    # version
    __canonical_name__ = "CustomApiEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    api_code: str
    func_name: str
    settings: dict[str, Any] | None = None
    view_access: bool = True
    helper_functions: dict[str, str] | None = None
    state: dict[Any, Any] | None = None
    signature: Signature

    __exclude_sync_diff_attrs__ = ["state"]

    def __repr__(self) -> str:
        type_name = type(self).__name__
        repr_str = f"""<{type_name}: {self.func_name}>

        {self.api_code}
        """
        return textwrap.dedent(repr_str)

    @field_validator("api_code", check_fields=False)
    @classmethod
    def validate_api_code(cls, api_code: str) -> str:
        valid_code = True
        api_code = dedent(api_code)
        try:
            ast.parse(api_code)
        except SyntaxError:
            # If the code isn't valid Python syntax
            valid_code = False

        if not valid_code:
            raise ValueError("Code must be a valid Python function.")

        return api_code

    @field_validator("func_name", check_fields=False)
    @classmethod
    def validate_func_name(cls, func_name: str) -> str:
        if not str.isidentifier(func_name) or keyword.iskeyword(func_name):
            raise ValueError("Invalid function name.")
        return func_name

    @field_validator("settings", check_fields=False)
    @classmethod
    def validate_settings(
        cls, settings: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        return settings

    def update_state(self, state: dict[Any, Any]) -> None:
        self.state = state

    def build_internal_context(
        self,
        context: AuthedServiceContext,
        admin_client: SyftClient | None = None,
        user_client: SyftClient | None = None,
    ) -> TwinAPIAuthedContext:
        helper_function_dict: dict[str, Callable] = {}
        self.helper_functions = self.helper_functions or {}
        for helper_name, helper_code in self.helper_functions.items():
            # Create a dictionary to serve as local scope
            local_scope: dict[str, Callable] = {}

            # Execute the function string within the local scope
            exec(helper_code, local_scope)  # nosec
            helper_function_dict[helper_name] = local_scope[helper_name]

        helper_function_set = HelperFunctionSet(helper_function_dict)

        user = context.server.services.user.get_current_user(context)

        return TwinAPIAuthedContext(
            credentials=context.credentials,
            role=context.role,
            job_id=context.job_id,
            extra_kwargs=context.extra_kwargs,
            has_execute_permissions=context.has_execute_permissions,
            server=context.server,
            id=context.id,
            settings=self.settings or {},
            code=helper_function_set,
            state=self.state or {},
            user=user,
            admin_client=admin_client,
            user_client=user_client,
        )

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        # relative
        from ... import SyftSigningKey
        from ..context import AuthedServiceContext

        mock_context = AuthedServiceContext(
            server=AbstractServer(), credentials=SyftSigningKey.generate().verify_key
        )
        return self.call_locally(mock_context, *args, **kwargs)

    def call_locally(
        self, context: AuthedServiceContext, *args: Any, **kwargs: Any
    ) -> Any:
        inner_function = ast.parse(self.api_code).body[0]
        inner_function.decorator_list = []
        # compile the function
        raw_byte_code = compile(ast.unparse(inner_function), "<string>", "exec")

        # load it
        exec(raw_byte_code)  # nosec

        internal_context = self.build_internal_context(context=context)

        # execute it
        evil_string = f"{self.func_name}(*args, **kwargs,context=internal_context)"
        result = eval(evil_string, None, locals())  # nosec

        # Update code context state
        self.update_state(internal_context.state)

        # return the results
        return result


@serializable()
class PrivateAPIEndpoint(Endpoint):
    # version
    __canonical_name__ = "PrivateAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    view_access: bool = False


@serializable()
class PublicAPIEndpoint(Endpoint):
    # version
    __canonical_name__ = "PublicAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1


class BaseTwinAPIEndpoint(SyftObject):
    __canonical_name__ = "BaseTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    @model_validator(mode="before")
    @classmethod
    def validate_signature(cls, data: dict[str, Any]) -> dict[str, Any]:
        mock_function = data["mock_function"]  # mock_function can't be None
        private_function = data.get("private_function")

        # Add none check
        if private_function and private_function.signature != mock_function.signature:
            raise ValueError(
                "Mock and Private API Endpoints must have the same signature."
            )

        return data

    @field_validator("path", check_fields=False)
    @classmethod
    def validate_path(cls, path: str) -> str:
        # TODO: Check path doesn't collide with system endpoints

        if path.startswith(".") or path.endswith("."):
            raise ValueError("Path cannot start or end with a '.'")
        if not path.islower():
            raise ValueError("Path must be lowercase")
        parts = path.split(".")
        for part in parts:
            if not str.isidentifier(part) or keyword.iskeyword(part):
                raise ValueError(f"Invalid path: {part} is not a valid identifier")

        return path

    @field_validator("private_function", check_fields=False)
    @classmethod
    def validate_private_function(
        cls, private_function: PrivateAPIEndpoint | None
    ) -> PrivateAPIEndpoint | None:
        # TODO: what kind of validation should we do here?

        return private_function

    @field_validator("mock_function", check_fields=False)
    @classmethod
    def validate_mock_function(
        cls, mock_function: PublicAPIEndpoint
    ) -> PublicAPIEndpoint:
        # TODO: what kind of validation should we do here?
        return mock_function


@serializable()
class UpdateTwinAPIEndpoint(PartialSyftObject, BaseTwinAPIEndpoint):
    # version
    __canonical_name__ = "UpdateTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    description: MarkdownDescription | None = None
    endpoint_timeout: int = 60


@serializable()
class CreateTwinAPIEndpointV1(BaseTwinAPIEndpoint):
    # version
    __canonical_name__ = "CreateTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    signature: Signature
    description: MarkdownDescription | None = None
    worker_pool: str | None = None
    endpoint_timeout: int = 60


@serializable()
class CreateTwinAPIEndpoint(BaseTwinAPIEndpoint):
    # version
    __canonical_name__ = "CreateTwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_2

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    signature: Signature
    description: MarkdownDescription | None = None
    worker_pool_name: str | None = None
    endpoint_timeout: int = 60

    def __init__(
        self, description: str | MarkdownDescription | None = "", **kwargs: Any
    ) -> None:
        if isinstance(description, str):
            description = MarkdownDescription(text=description)

        super().__init__(**kwargs, description=description)


@serializable()
class TwinAPIEndpointV1(SyncableSyftObject):
    # version
    __canonical_name__: str = "TwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_1
    __exclude_sync_diff_attrs__ = ["private_function"]
    __private_sync_attr_mocks__ = {
        "private_function": None,
    }

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    signature: Signature
    description: MarkdownDescription | None = None
    action_object_id: UID
    worker_pool: str | None = None
    endpoint_timeout: int = 60

    __attr_searchable__ = ["path"]
    __attr_unique__ = ["path"]
    __repr_attrs__ = [
        "path",
        "description",
        "private_function",
        "mock_function",
        "endpoint_timeout",
    ]


@serializable()
class TwinAPIEndpoint(SyncableSyftObject):
    # version
    __canonical_name__: str = "TwinAPIEndpoint"
    __version__ = SYFT_OBJECT_VERSION_2
    __exclude_sync_diff_attrs__ = ["private_function"]
    __private_sync_attr_mocks__ = {
        "private_function": None,
    }

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    path: str
    private_function: PrivateAPIEndpoint | None = None
    mock_function: PublicAPIEndpoint
    signature: Signature
    description: MarkdownDescription | None = None
    action_object_id: UID
    worker_pool_name: str | None = None
    endpoint_timeout: int = 60

    __attr_searchable__ = ["path"]
    __attr_unique__ = ["path"]
    __repr_attrs__ = [
        "path",
        "description",
        "private_function",
        "mock_function",
        "endpoint_timeout",
    ]

    def has_mock(self) -> bool:
        return self.api_mock_code is not None

    def has_permission(self, context: AuthedServiceContext) -> bool:
        """Check if the user has permission to access the endpoint.

        Args:
            context: The context of the user requesting the code.
        Returns:
            bool: True if the user has permission to access the endpoint, False otherwise.
        """
        if context.role.value == 128:
            return True
        return False

    def select_code(
        self, context: AuthedServiceContext
    ) -> PrivateAPIEndpoint | PublicAPIEndpoint | None:
        """Select the code to execute based on the user's permissions and public code availability.

        Args:
            context: The context of the user requesting the code.
        Returns:
            Result[Ok, Err]: The selected code to execute.
        """
        if self.has_permission(context) and self.private_function:
            return self.private_function
        return self.mock_function

    def exec(
        self,
        context: AuthedServiceContext,
        *args: Any,
        log_id: UID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Execute the code based on the user's permissions and public code availability.

        Args:
            context: The context of the user requesting the code.
            *args: Any
            **kwargs: Any
        Returns:
            Any: The result of the executed code.
        """
        selected_code = self.select_code(context)
        return self.exec_code(selected_code, context, *args, log_id=log_id, **kwargs)

    def exec_mock_function(
        self,
        context: AuthedServiceContext,
        *args: Any,
        log_id: UID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Execute the public code if it exists."""
        if self.mock_function:
            return self.exec_code(
                self.mock_function, context, *args, log_id=log_id, **kwargs
            )

        raise SyftException(public_message="No public code available")

    def exec_private_function(
        self,
        context: AuthedServiceContext,
        *args: Any,
        log_id: UID | None = None,
        **kwargs: Any,
    ) -> Any:
        """Execute the private code if user is has the proper permissions.

        Args:
            context: The context of the user requesting the code.
            *args: Any
            **kwargs: Any
        Returns:
            Any: The result of the executed code.
        """
        if self.private_function is None:
            raise SyftException(public_message="No private code available")

        if self.has_permission(context):
            return self.exec_code(
                self.private_function, context, *args, log_id=log_id, **kwargs
            )

        raise SyftException(public_message="You're not allowed to run this code.")

    def get_user_client_from_server(self, context: AuthedServiceContext) -> SyftClient:
        # get a user client
        guest_client = context.server.get_guest_client()
        user_client = guest_client
        private_key = context.server.services.user.signing_key_for_verify_key(
            context.credentials
        )
        signing_key = private_key.signing_key
        user_client.credentials = signing_key
        return user_client

    def get_admin_client_from_server(self, context: AuthedServiceContext) -> SyftClient:
        admin_client = context.server.get_guest_client()
        admin_client.credentials = context.server.signing_key
        return admin_client

    @as_result(SyftException)
    def exec_code(
        self,
        code: PrivateAPIEndpoint | PublicAPIEndpoint,
        context: AuthedServiceContext,
        *args: Any,
        log_id: UID | None = None,
        **kwargs: Any,
    ) -> Any:
        # stdlib
        import builtins as __builtin__
        import functools

        original_print = __builtin__.print

        try:
            if log_id is not None:
                print = functools.partial(log_print, context, log_id)
            else:
                print = original_print  # type: ignore

            inner_function = ast.parse(code.api_code).body[0]
            inner_function.decorator_list = []
            # compile the function
            src = ast.unparse(inner_function)
            raw_byte_code = compile(src, code.func_name, "exec")
            register_fn_in_linecache(code.func_name, src)
            user_client = self.get_user_client_from_server(context)
            admin_client = self.get_admin_client_from_server(context)

            internal_context = code.build_internal_context(
                context=context, admin_client=admin_client, user_client=user_client
            )
            evil_string = f"{code.func_name}(*args, **kwargs,context=internal_context)"

            _globals = {"print": print}
            # load it
            exec(raw_byte_code, _globals, locals())  # nosec

            # execute it
            evil_string = f"{code.func_name}(*args, **kwargs,context=internal_context)"
            result = None
            try:
                # users can raise SyftException in their code
                result = eval(evil_string, _globals, locals())  # nosec
            except SyftException as e:
                # capture it as the result variable
                result = e

            # run all this code to clean up the state
            code.update_state(internal_context.state)

            if isinstance(code, PublicAPIEndpoint):
                self.mock_function = code
            else:
                self.private_function = code  # type: ignore

            api_service = context.server.get_service("apiservice")
            api_service.stash.upsert(
                context.server.services.user.root_verify_key, self
            ).unwrap()

            print = original_print  # type: ignore
            # if we caught a SyftException above we will raise and auto wrap to Result
            if isinstance(result, SyftException):
                raise result

            # here we got a non Exception result which will also be wrapped in Result
            # return the results
            return result
        except Exception as e:
            # If it's admin, return the error message.
            # TODO: cleanup typeerrors
            if context.role.value == 128 or isinstance(e, TypeError):
                raise SyftException(
                    public_message=f"An error was raised during the execution of the API endpoint call: \n {str(e)}"
                )
            else:
                raise SyftException(
                    public_message=(
                        "Oops something went wrong during this endpoint execution, "
                        "please contact your admin."
                    )
                )


def set_access_type(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        if context.obj.private_function is not None:
            context.output["access"] = "Private / Mock"
        else:
            context.output["access"] = "Public"
    return context


def check_and_cleanup_signature(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        params = dict(context.obj.signature.parameters)
        if "context" not in params:
            raise ValueError(
                "Function Signature must include 'context' [AuthedContext] parameters."
            )
        params.pop("context", None)
        context.output["signature"] = Signature(
            list(params.values()),
            return_annotation=context.obj.signature.return_annotation,
        )
    return context


def decorator_cleanup(code: str) -> str:
    # Regular expression to remove decorator
    # It matches from "@" to "def" (non-greedy) across multiple lines
    decorator_regex = r"@.*?def "

    # Substituting the matched pattern with "def"
    return re.sub(decorator_regex, "def ", code, count=1, flags=re.DOTALL)


def extract_code_string(code_field: str) -> Callable:
    def code_string(context: TransformContext) -> TransformContext:
        if context.obj is not None and context.output is not None:
            endpoint_type = (
                context.obj.private_function
                if code_field == "private_function"
                else context.obj.mock_function
            )
            helper_function_field = (
                "mock_helper_functions"
                if code_field == "mock_function"
                else "private_helper_functions"
            )

            context.server = cast(AbstractServer, context.server)
            admin_key = context.server.services.user.root_verify_key

            # If endpoint exists **AND** (has visible access **OR** the user is admin)
            if endpoint_type is not None and (
                endpoint_type.view_access or context.credentials == admin_key
            ):
                context.output[code_field] = decorator_cleanup(endpoint_type.api_code)
                context.output[helper_function_field] = (
                    endpoint_type.helper_functions.values() or []
                )
            else:
                context.output[code_field] = None
                context.output[helper_function_field] = []
        return context

    return code_string


@transform(TwinAPIAuthedContext, TwinAPIContextView)
def twin_api_context_to_twin_api_context_view() -> list[Callable]:
    return [keep(["state", "settings", "user"])]


@transform(CreateTwinAPIEndpoint, TwinAPIEndpoint)
def endpoint_create_to_twin_endpoint() -> list[Callable]:
    return [generate_id, generate_action_object_id, check_and_cleanup_signature]


@transform(TwinAPIEndpoint, TwinAPIEndpointView)
def twin_endpoint_to_view() -> list[Callable]:
    return [
        set_access_type,
        extract_code_string("private_function"),
        extract_code_string("mock_function"),
    ]


@transform(Endpoint, PrivateAPIEndpoint)
def endpoint_to_private_endpoint() -> list[Callable]:
    return [
        keep(
            [
                "api_code",
                "func_name",
                "settings",
                "view_access",
                "helper_functions",
                "state",
                "signature",
            ]
        )
    ]


@transform(Endpoint, PublicAPIEndpoint)
def endpoint_to_public_endpoint() -> list[Callable]:
    return [
        keep(
            [
                "api_code",
                "func_name",
                "settings",
                "view_access",
                "helper_functions",
                "state",
                "signature",
            ]
        )
    ]


@migrate(TwinAPIEndpointV1, TwinAPIEndpoint)
def migrate_twin_api_endpoint_v1_to_current() -> list[Callable]:
    return [
        drop(["worker_pool"]),
        make_set_default("worker_pool_name", None),
    ]


@migrate(CreateTwinAPIEndpointV1, CreateTwinAPIEndpoint)
def migrate_create_twin_api_endpoint_v1_to_current() -> list[Callable]:
    return [
        drop(["worker_pool"]),
        make_set_default("worker_pool_name", None),
    ]


@migrate(TwinAPIEndpointViewV1, TwinAPIEndpointView)
def migrate_twin_api_endpoint_view_v1_to_current() -> list[Callable]:
    return [
        drop(["worker_pool"]),
        make_set_default("worker_pool_name", None),
    ]


@migrate(TwinAPIEndpointView, TwinAPIEndpointViewV1)
def migrate_twin_api_endpoint_view_current_to_v1() -> list[Callable]:
    return [
        drop(["worker_pool_name"]),
        make_set_default("worker_pool", None),
    ]


@migrate(CreateTwinAPIEndpoint, CreateTwinAPIEndpointV1)
def migrate_create_twin_api_endpoint_current_to_v1() -> list[Callable]:
    return [
        drop(["worker_pool_name"]),
        make_set_default("worker_pool", None),
    ]


@migrate(TwinAPIEndpoint, TwinAPIEndpointV1)
def migrate_twin_api_endpoint_current_to_v1() -> list[Callable]:
    return [
        drop(["worker_pool_name"]),
        make_set_default("worker_pool", None),
    ]


def api_endpoint(
    path: str,
    settings: dict[str, str] | None = None,
    helper_functions: list[Callable] | None = None,
    description: MarkdownDescription | None = None,
    worker_pool_name: str | None = None,
    endpoint_timeout: int = 60,
) -> Callable[..., TwinAPIEndpoint | SyftError]:
    def decorator(f: Callable) -> TwinAPIEndpoint | SyftError:
        try:
            helper_functions_dict = {
                f.__name__: dedent(inspect.getsource(f))
                for f in (helper_functions or [])
            }
            res = CreateTwinAPIEndpoint(
                path=path,
                mock_function=PublicAPIEndpoint(
                    api_code=inspect.getsource(f),
                    func_name=f.__name__,
                    settings=settings,
                    signature=inspect.signature(f),
                    helper_functions=helper_functions_dict,
                ),
                signature=inspect.signature(f),
                description=description,
                worker_pool_name=worker_pool_name,
                endpoint_timeout=endpoint_timeout,
            )
        except ValidationError as e:
            for error in e.errors():
                error_msg = error["msg"]
            res = SyftError(message=error_msg)
        return res

    return decorator


def api_endpoint_method(
    settings: dict[str, str] | None = None,
    helper_functions: list[Callable] | None = None,
) -> Callable[..., Endpoint | SyftError]:
    def decorator(f: Callable) -> Endpoint | SyftError:
        try:
            helper_functions_dict = {
                f.__name__: dedent(inspect.getsource(f))
                for f in (helper_functions or [])
            }
            return Endpoint(
                api_code=inspect.getsource(f),
                func_name=f.__name__,
                settings=settings,
                signature=inspect.signature(f),
                helper_functions=helper_functions_dict,
            )
        except ValidationError as e:
            for error in e.errors():
                error_msg = error["msg"]
            res = SyftError(message=error_msg)
        return res

    return decorator


def create_new_api_endpoint(
    path: str,
    mock_function: Endpoint,
    private_function: Endpoint | None = None,
    description: MarkdownDescription | None = None,
    worker_pool_name: str | None = None,
    endpoint_timeout: int = 60,
    hide_mock_definition: bool = False,
    hide_private_definition: bool = True,
) -> CreateTwinAPIEndpoint | SyftError:
    try:
        # Parse the string to extract the function name

        endpoint_signature = mock_function.signature
        if private_function is not None:
            if private_function.signature != mock_function.signature:
                return SyftError(message="Signatures don't match")
            endpoint_signature = mock_function.signature
            private_function.view_access = not hide_private_definition
            mock_function.view_access = not hide_mock_definition

            return CreateTwinAPIEndpoint(
                path=path,
                private_function=private_function.to(PrivateAPIEndpoint),
                mock_function=mock_function.to(PublicAPIEndpoint),
                signature=endpoint_signature,
                description=description,
                worker_pool_name=worker_pool_name,
                endpoint_timeout=endpoint_timeout,
            )

        return CreateTwinAPIEndpoint(
            path=path,
            prublic_code=mock_function.to(PublicAPIEndpoint),
            signature=endpoint_signature,
            worker_pool_name=worker_pool_name,
            endpoint_timeout=endpoint_timeout,
        )
    except ValidationError as e:
        for error in e.errors():
            error_msg = error["msg"]

    return SyftError(message=error_msg)



# stdlib

# third party

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from ...util.telemetry import instrument
from .notifier import NotifierSettings


@instrument
@serializable(canonical_name="NotifierSQLStash", version=1)
class NotifierStash(ObjectStash[NotifierSettings]):
    @as_result(StashException, NotFoundException)
    def get(self, credentials: SyftVerifyKey) -> NotifierSettings:
        """Get Settings"""
        # actually get latest settings
        result = self.get_all(credentials, limit=1, sort_order="desc").unwrap()
        if len(result) > 0:
            return result[0]
        raise NotFoundException(
            public_message="No settings found for the current user."
        )



# future
from __future__ import annotations

# third party
from pydantic import BaseModel
from pydantic import Field

# relative
from ..serde.serializable import serializable
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from .locks import LockingConfig
from .locks import NoLockingConfig


class StoreClientConfig(BaseModel):
    """Base Client specific configuration"""

    pass


@serializable(
    attrs=["settings", "store_config", "unique_cks", "searchable_cks"],
    canonical_name="StorePartition",
    version=1,
)
class StorePartition:
    """Base StorePartition
    Parameters:
        settings: PartitionSettings
            PySyft specific settings
        store_config: StoreConfig
            Backend specific configuration
    """


@serializable()
class StoreConfig(SyftBaseObject):
    """Base Store configuration

    Parameters:
        store_type: Type
            Document Store type
        client_config: Optional[StoreClientConfig]
            Backend-specific config
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
            Defaults to NoLockingConfig.
    """

    __canonical_name__ = "StoreConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    store_type: type[DocumentStore]
    client_config: StoreClientConfig | None = None
    locking_config: LockingConfig = Field(default_factory=NoLockingConfig)  # noqa: F821


@serializable(canonical_name="DocumentStore", version=1)
class DocumentStore:
    pass



# stdlib

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .policy import UserPolicy


@serializable(canonical_name="UserPolicySQLStash", version=1)
class UserPolicyStash(ObjectStash[UserPolicy]):
    @as_result(StashException, NotFoundException)
    def get_all_by_user_verify_key(
        self, credentials: SyftVerifyKey, user_verify_key: SyftVerifyKey
    ) -> list[UserPolicy]:
        return self.get_all(
            credentials=credentials,
            filters={"user_verify_key": user_verify_key},
        ).unwrap()



# stdlib

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ..service import AbstractService
from .queue_stash import QueueStash


@serializable(canonical_name="QueueService", version=1)
class QueueService(AbstractService):
    stash: QueueStash

    def __init__(self, store: DBManager) -> None:
        self.stash = QueueStash(store=store)



# third party
from tqdm import tqdm

# syft absolute
import syft as sy

# relative
from ...client.datasite_client import DatasiteClient
from ...client.syncing import compare_clients
from ...service.code.user_code import UserCode
from ...service.job.job_stash import Job
from ...service.job.job_stash import JobStatus
from ...service.request.request import Request
from ...service.request.request import RequestStatus
from ...service.sync.diff_state import ObjectDiffBatch
from ...types.result import Err


def deny_requests_without_autosync_tag(client_low: DatasiteClient):
    # Deny all requests that are not autosync
    requests = client_low.requests.get_all()
    if isinstance(requests, sy.SyftError):
        print(requests)
        return

    denied_requests = []
    for request in tqdm(requests):
        if request.status != RequestStatus.PENDING:
            continue
        if "autosync" not in request.tags:
            request.deny(
                reason="This request has been denied automatically. "
                "Please use the designated API to submit your request."
            )
            denied_requests.append(request.id)
    print(f"Denied {len(denied_requests)} requests without autosync tag")


def is_request_to_sync(batch: ObjectDiffBatch) -> bool:
    # True if this is a new low-side request
    # TODO add condition for sql requests/usercodes
    low_request = batch.root.low_obj
    return (
        isinstance(low_request, Request)
        and batch.status == "NEW"
        and "autosync" in low_request.tags
    )


def is_job_to_sync(batch: ObjectDiffBatch):
    # True if this is a new high-side job that is either COMPLETED or ERRORED
    if batch.status != "NEW":
        return False
    if not isinstance(batch.root.high_obj, Job):
        return False
    job = batch.root.high_obj
    return job.status in (JobStatus.ERRORED, JobStatus.COMPLETED)


def execute_requests(
    client_high: DatasiteClient, request_ids: list[sy.UID]
) -> dict[sy.UID, Job]:
    jobs_by_request_id = {}
    for request_id in request_ids:
        request = client_high.requests.get_by_uid(request_id)
        if not isinstance(request, Request):
            continue

        code = request.code
        if not isinstance(code, UserCode):
            continue

        func_name = request.code.service_func_name
        api_func = getattr(client_high.code, func_name, None)
        if api_func is None:
            continue

        job = api_func(blocking=False)
        jobs_by_request_id[request_id] = job

    return jobs_by_request_id


def deny_failed_jobs(
    client_low: DatasiteClient,
    jobs: list[Job],
) -> None:
    # NOTE no syncing is needed, requests are denied on the low side
    denied_requests = []

    for job in jobs:
        if job.status != JobStatus.ERRORED:
            continue

        error_result = job.result
        if isinstance(error_result, Err):
            error_msg = error_result.err_value
        else:
            error_msg = "An unknown error occurred, please check the Job logs for more information."

        code_id = job.user_code_id
        if code_id is None:
            continue
        requests = client_low.requests.get_by_usercode_id(code_id)
        if isinstance(requests, list) and len(requests) > 0:
            request = requests[0]
            request.deny(reason=f"Execution failed: {error_msg}")
            denied_requests.append(request.id)
        else:
            print(f"Failed to deny request for job {job.id}")

    print(f"Denied {len(denied_requests)} failed requests")


def sync_finished_jobs(
    client_low: DatasiteClient,
    client_high: DatasiteClient,
) -> dict[sy.UID, sy.SyftError | sy.SyftSuccess] | sy.SyftError:
    sync_job_results = {}
    synced_jobs = []
    diff = compare_clients(
        from_client=client_high, to_client=client_low, include_types=["job"]
    )
    if isinstance(diff, sy.SyftError):
        print(diff)
        return diff

    for batch in diff.batches:
        if is_job_to_sync(batch):
            job = batch.root.high_obj

            w = batch.resolve(build_state=False)
            share_result = w.click_share_all_private_data()
            if isinstance(share_result, sy.SyftError):
                sync_job_results[job.id] = share_result
                continue
            sync_result = w.click_sync()

            synced_jobs.append(job)
            sync_job_results[job.id] = sync_result

    print(f"Sharing {len(sync_job_results)} new results")
    deny_failed_jobs(client_low, synced_jobs)
    return sync_job_results


def sync_new_requests(
    client_low: DatasiteClient,
    client_high: DatasiteClient,
) -> dict[sy.UID, sy.SyftSuccess | sy.SyftError] | sy.SyftError:
    sync_request_results = {}
    diff = compare_clients(
        from_client=client_low, to_client=client_high, include_types=["request"]
    )
    if isinstance(diff, sy.SyftError):
        print(diff)
        return sync_request_results
    print(f"{len(diff.batches)} request batches found")
    for batch in tqdm(diff.batches):
        if is_request_to_sync(batch):
            request_id = batch.root.low_obj.id
            w = batch.resolve(build_state=False)
            result = w.click_sync()
            sync_request_results[request_id] = result
    return sync_request_results


def sync_and_execute_new_requests(
    client_low: DatasiteClient, client_high: DatasiteClient
) -> None:
    sync_results = sync_new_requests(client_low, client_high)
    if isinstance(sync_results, sy.SyftError):
        print(sync_results)
        return

    request_ids = [
        uid for uid, res in sync_results.items() if isinstance(res, sy.SyftSuccess)
    ]
    print(f"Synced {len(request_ids)} new requests")

    jobs_by_request = execute_requests(client_high, request_ids)
    print(f"Started {len(jobs_by_request)} new jobs")


def auto_sync(client_low: DatasiteClient, client_high: DatasiteClient) -> None:
    print("Starting auto sync")
    print("Denying non tagged jobs")
    deny_requests_without_autosync_tag(client_low)
    print("Syncing and executing")
    sync_and_execute_new_requests(client_low, client_high)
    sync_finished_jobs(client_low, client_high)
    print("Finished auto sync")



# stdlib
from collections.abc import Callable

# syft absolute
import syft as sy

# relative
from ..... import test_settings
from ..rate_limiter import is_within_rate_limit


def make_test_query(settings) -> Callable:
    updated_settings = {
        "calls_per_min": 10,
        "rate_limiter_enabled": True,
        "credentials": test_settings.gce_service_account.to_dict(),
        "region": test_settings.gce_region,
        "project_id": test_settings.gce_project_id,
    } | settings

    # these are the same if you allow the rate limiter to be turned on and off
    @sy.api_endpoint_method(
        settings=updated_settings,
        helper_functions=[is_within_rate_limit],
    )
    def live_test_query(
        context,
        sql_query: str,
    ) -> str:
        # stdlib
        import datetime

        # third party
        from google.cloud import bigquery  # noqa: F811
        from google.oauth2 import service_account

        # syft absolute
        from syft import SyftException

        # Auth for Bigquer based on the workload identity
        credentials = service_account.Credentials.from_service_account_info(
            context.settings["credentials"]
        )
        scoped_credentials = credentials.with_scopes(
            ["https://www.googleapis.com/auth/cloud-platform"]
        )

        client = bigquery.Client(
            credentials=scoped_credentials,
            location=context.settings["region"],
        )

        # Store a dict with the calltimes for each user, via the email.
        if context.settings["rate_limiter_enabled"]:
            if context.user.email not in context.state.keys():
                context.state[context.user.email] = []

            if not context.code.is_within_rate_limit(context):
                raise SyftException(
                    public_message="Rate limit of calls per minute has been reached."
                )
            context.state[context.user.email].append(datetime.datetime.now())

        try:
            rows = client.query_and_wait(
                sql_query,
                project=context.settings["project_id"],
            )

            if rows.total_rows > 1_000_000:
                raise SyftException(
                    public_message="Please only write queries that gather aggregate statistics"
                )

            return rows.to_dataframe()

        except Exception as e:
            # not a bigquery exception
            if not hasattr(e, "_errors"):
                output = f"got exception e: {type(e)} {str(e)}"
                raise SyftException(
                    public_message=f"An error occured executing the API call {output}"
                )

            # Treat all errors that we would like to be forwarded to the data scientists
            # By default, any exception is only visible to the data owner.

            if e._errors[0]["reason"] in [
                "badRequest",
                "blocked",
                "duplicate",
                "invalidQuery",
                "invalid",
                "jobBackendError",
                "jobInternalError",
                "notFound",
                "notImplemented",
                "rateLimitExceeded",
                "resourceInUse",
                "resourcesExceeded",
                "tableUnavailable",
                "timeout",
            ]:
                raise SyftException(
                    public_message="Error occured during the call: "
                    + e._errors[0]["message"]
                )
            else:
                raise SyftException(
                    public_message="An error occured executing the API call, please contact the domain owner."
                )

    return live_test_query



# stdlib

# third party

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import StashException
from ...types.result import as_result
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from .data_subject import DataSubject
from .data_subject import DataSubjectCreate


@serializable(canonical_name="DataSubjectSQLStash", version=1)
class DataSubjectStash(ObjectStash[DataSubject]):
    @as_result(StashException)
    def get_by_name(self, credentials: SyftVerifyKey, name: str) -> DataSubject:
        return self.get_one(
            credentials=credentials,
            filters={"name": name},
        ).unwrap()


@serializable(canonical_name="DataSubjectService", version=1)
class DataSubjectService(AbstractService):
    stash: DataSubjectStash

    def __init__(self, store: DBManager) -> None:
        self.stash = DataSubjectStash(store=store)

    @service_method(path="data_subject.add", name="add_data_subject")
    def add(
        self, context: AuthedServiceContext, data_subject: DataSubjectCreate
    ) -> SyftSuccess:
        """Register a data subject."""  #

        member_relationships: set[tuple[str, str]] = data_subject.member_relationships
        if len(member_relationships) == 0:
            self.stash.set(
                context.credentials,
                data_subject.to(DataSubject, context=context),
            ).unwrap()
        else:
            for member_relationship in member_relationships:
                parent_ds, child_ds = member_relationship
                for ds in [parent_ds, child_ds]:
                    self.stash.set(
                        context.credentials,
                        ds.to(DataSubject, context=context),
                        ignore_duplicates=True,
                    ).unwrap()
                context.server.services.data_subject_member.add(
                    context, parent_ds.name, child_ds.name
                )

        return SyftSuccess(
            message=f"{len(member_relationships)+1} Data Subjects Registered",
            value=member_relationships,
        )

    @service_method(path="data_subject.get_all", name="get_all")
    def get_all(self, context: AuthedServiceContext) -> list[DataSubject]:
        """Get all Data subjects"""
        return self.stash.get_all(context.credentials).unwrap()

    @service_method(path="data_subject.get_members", name="members_for")
    def get_members(
        self, context: AuthedServiceContext, data_subject_name: str
    ) -> list[DataSubject]:
        relatives = context.server.services.data_subject.get_relatives(
            context, data_subject_name
        )

        members = []
        for relative in relatives:
            result = self.get_by_name(context=context, name=relative.child)
            members.append(result)

        return members

    @service_method(path="data_subject.get_by_name", name="get_by_name")
    def get_by_name(self, context: AuthedServiceContext, name: str) -> DataSubject:
        """Get a Data Subject by its name."""
        return self.stash.get_by_name(context.credentials, name=name).unwrap()


TYPE_TO_SERVICE[DataSubject] = DataSubjectService
SERVICE_TO_TYPES[DataSubjectService].update({DataSubject})



# stdlib
from collections.abc import Callable

# syft absolute
import syft as sy

# relative
from ..rate_limiter import is_within_rate_limit
from .data import schema_dict


def make_schema(
    settings,
    worker_pool_name,
    path="bigquery.schema",
) -> Callable:
    updated_settings = {
        "calls_per_min": 5,
        "rate_limiter_enabled": True,
        "schema_dict": schema_dict,
    } | settings

    @sy.api_endpoint(
        path=path,
        description="This endpoint allows for visualising the metadata of tables available in BigQuery.",
        settings=updated_settings,
        helper_functions=[is_within_rate_limit],
        worker_pool_name=worker_pool_name,
    )
    def mock_schema(
        context,
    ) -> str:
        # syft absolute
        from syft import SyftException

        # Store a dict with the calltimes for each user, via the email.
        if context.settings["rate_limiter_enabled"]:
            # stdlib
            import datetime

            if context.user.email not in context.state.keys():
                context.state[context.user.email] = []

            if not context.code.is_within_rate_limit(context):
                raise SyftException(
                    public_message="Rate limit of calls per minute has been reached."
                )
            context.state[context.user.email].append(datetime.datetime.now())

        # third party
        import pandas as pd

        df = pd.DataFrame(context.settings["schema_dict"])
        return df

    return mock_schema



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from .request import Request


@serializable(canonical_name="RequestStashSQL", version=1)
class RequestStash(ObjectStash[Request]):
    @as_result(SyftException)
    def get_all_for_verify_key(
        self,
        credentials: SyftVerifyKey,
        verify_key: SyftVerifyKey,
    ) -> list[Request]:
        return self.get_all(
            credentials=credentials,
            filters={"requesting_user_verify_key": verify_key},
        ).unwrap()

    @as_result(SyftException)
    def get_by_usercode_id(
        self, credentials: SyftVerifyKey, user_code_id: UID
    ) -> list[Request]:
        return self.get_all(
            credentials=credentials,
            filters={"code_id": user_code_id},
        ).unwrap()



# stdlib
from collections.abc import Callable

# syft absolute
import syft as sy

# relative
from ..rate_limiter import is_within_rate_limit
from .data import query_dict


def extract_limit_value(sql_query: str) -> int:
    # stdlib
    import re

    limit_pattern = re.compile(r"\bLIMIT\s+(\d+)\b", re.IGNORECASE)
    match = limit_pattern.search(sql_query)
    if match:
        return int(match.group(1))
    return None


def is_valid_sql(query: str) -> bool:
    # stdlib
    import sqlite3

    # Prepare an in-memory SQLite database
    conn = sqlite3.connect(":memory:")
    cursor = conn.cursor()

    try:
        # Use the EXPLAIN QUERY PLAN command to get the query plan
        cursor.execute(f"EXPLAIN QUERY PLAN {query}")
    except sqlite3.Error as e:
        if "no such table" in str(e).lower():
            return True
        return False
    finally:
        conn.close()


def adjust_dataframe_rows(df, target_rows: int):
    # third party
    import pandas as pd

    current_rows = len(df)

    if target_rows > current_rows:
        # Repeat rows to match target_rows
        repeat_times = (target_rows + current_rows - 1) // current_rows
        df_expanded = pd.concat([df] * repeat_times, ignore_index=True).head(
            target_rows
        )
    else:
        # Truncate rows to match target_rows
        df_expanded = df.head(target_rows)

    return df_expanded


def make_test_query(settings: dict) -> Callable:
    updated_settings = {
        "calls_per_min": 10,
        "rate_limiter_enabled": True,
        "query_dict": query_dict,
    } | settings

    # these are the same if you allow the rate limiter to be turned on and off
    @sy.api_endpoint_method(
        settings=updated_settings,
        helper_functions=[
            is_within_rate_limit,
            extract_limit_value,
            is_valid_sql,
            adjust_dataframe_rows,
        ],
    )
    def mock_test_query(
        context,
        sql_query: str,
    ) -> str:
        # stdlib
        import datetime

        # third party
        from google.api_core.exceptions import BadRequest

        # syft absolute
        from syft import SyftException

        # Store a dict with the calltimes for each user, via the email.
        if context.settings["rate_limiter_enabled"]:
            if context.user.email not in context.state.keys():
                context.state[context.user.email] = []

            if not context.code.is_within_rate_limit(context):
                raise SyftException(
                    public_message="Rate limit of calls per minute has been reached."
                )
            context.state[context.user.email].append(datetime.datetime.now())

        bad_table = "invalid_table"
        bad_post = (
            "BadRequest: 400 POST "
            "https://bigquery.googleapis.com/bigquery/v2/projects/project-id/"
            "queries?prettyPrint=false: "
        )
        if bad_table in sql_query:
            try:
                raise BadRequest(
                    f'{bad_post} Table "{bad_table}" must be qualified '
                    "with a dataset (e.g. dataset.table)."
                )
            except Exception as e:
                raise SyftException(
                    public_message=f"*must be qualified with a dataset*. {e}"
                )

        if not context.code.is_valid_sql(sql_query):
            raise BadRequest(
                f'{bad_post} Syntax error: Unexpected identifier "{sql_query}" at [1:1]'
            )

        # third party
        import pandas as pd

        limit = context.code.extract_limit_value(sql_query)
        if limit > 1_000_000:
            raise SyftException(
                public_message="Please only write queries that gather aggregate statistics"
            )

        base_df = pd.DataFrame(context.settings["query_dict"])

        df = context.code.adjust_dataframe_rows(base_df, limit)
        return df

    return mock_test_query
